#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#define MUNGPRIVATE
#include "mungwall.h"

#define WALLSIZE	40

struct memchunk {
	struct memchunk *next;
	void *memory;	/* size+2*WALLSIZE bytes from here, or NULL */
	long size;
};

static struct memchunk *memorylist;

/* Check if the wall is still untouched
 * return value: 0 if chunk is OK, 1 if it is damaged */
static int checkchunk(struct memchunk *chunk) {
	int i;
	int bad = 0;

	for (i = sizeof(long); i < WALLSIZE; i++) {
		if (((char *)chunk->memory)[i] != i) {
			fprintf(stderr,"checkmem(): memory %p bad in"
				" front wall, %dth byte\n",
				(char *)chunk->memory+WALLSIZE,i);
			bad = 1;
		}
	}
	for (i = 0; i < WALLSIZE; i++) {
		if (((char *)chunk->memory)[i+chunk->size+WALLSIZE] != WALLSIZE-i-1) {
			fprintf(stderr,"checkmem(): memory %p bad in"
				" back wall, %dth byte\n",
				(char *)chunk->memory+WALLSIZE,i);
			bad = 1;
		}
	}
	return bad;
}

void mung_checkmem(int line,char *file,int verbose) {
	struct memchunk *chunk;
	int bad = 0;

#ifdef VERBOSE
	if (verbose)
		fprintf(stderr,"%s:%d checkmem()\n",file,line);
#endif
	for (chunk = memorylist; chunk; chunk = chunk->next) {
		if (!chunk->memory) continue;
		if (*(long *)chunk->memory != chunk->size) {
			fprintf(stderr,"checkmem(): memory %p bad in first 4 bytes"
				" of front wall\n",(char *)chunk->memory+WALLSIZE);
				bad = 1;
		}
		if (checkchunk(chunk)) bad = 1;
	}
	if (bad) {
		fprintf(stderr,"checkmem() bailing out at %s:%d\n",file,line);
		abort();
	}
}

static void addchunk(void *memory,long size)
{
	struct memchunk *chunk = malloc(sizeof(struct memchunk));
	chunk->next = memorylist;
	chunk->memory = memory;
	chunk->size = size;
	memorylist = chunk;
}

static struct memchunk *findchunk(void *memory)
{
	struct memchunk *walk;
	for (walk = memorylist; walk; walk = walk->next)
		if (walk->memory == memory) return walk;
	return NULL;
}

void *mung_malloc(size_t size,int line,char *file)
{
	char *mem = malloc(size + 2*WALLSIZE);
	int i;

	for (i=0; i<WALLSIZE; i++) {
		mem[i] = i;
		mem[i+size+WALLSIZE] = WALLSIZE-i-1;
	}
	*((long *)mem) = size;
	addchunk(mem,size);
	mung_checkmem(line,file,0);
#ifdef VERBOSE
	fprintf(stderr,"%s:%d malloc(%d)=%p\n",file,line,size,mem+WALLSIZE);
#endif
	return mem + WALLSIZE;
}

void *mung_realloc(void *ptr,size_t size,int line,char *file)
{
	char *result;
	int i;
	struct memchunk *chunk = findchunk((char *)ptr - WALLSIZE);

	if (!chunk) {
		fprintf(stderr,"realloc(%p,%ld) at %s:%d mem not in memory list\n",
			ptr,(long)size,file,line);
		abort();
	}
	mung_checkmem(line,file,0);
	result = realloc((char *)ptr - WALLSIZE,size + 2*WALLSIZE);
	for (i=0; i<WALLSIZE; i++) result[i+size+WALLSIZE] = WALLSIZE-i-1;
	*((long *)result) = size;
	chunk->memory = result;
	chunk->size = size;
#ifdef VERBOSE
	fprintf(stderr,"%s:%d realloc(%p,%d)=%p\n",file,line,ptr,size,result+WALLSIZE);
#endif
	return result + WALLSIZE;
}

void *mung_calloc(size_t nmemb,size_t size,int line,char *file)
{
	char *mem;

	size *= nmemb;
	mem = mung_malloc(size,line,file);
	do mem[--size]=0; while (size>0);
#ifdef VERBOSE
	fprintf(stderr,"%s:%d calloc(%d)=%p\n",file,line,size,mem);
#endif
	return mem;
}

void mung_free(void *ptr,int line,char *file)
{
	struct memchunk *chunk;

#ifdef VERBOSE
	fprintf(stderr,"%s:%d free(%p)\n",file,line,ptr);
#endif
	chunk = findchunk((char *)ptr - WALLSIZE);
	if (!chunk) {
		fprintf(stderr,"free(%p) at %s:%d mem not in memory list\n",
			ptr,file,line);
		abort();
	}
	mung_checkmem(line,file,0);
	chunk->memory = NULL;
	free((char *)ptr - WALLSIZE);
}
